from abc import abstractmethod
from typing import Generic, Type, TypeVar
from pydantic.main import BaseModel
from fakts.beacon.beacon import FaktsEndpoint
import webbrowser
import asyncio
import uuid
from aiohttp import web
import aiohttp_cors
from urllib.parse import quote, urlencode
from koil import koil


class RetrieverException(Exception):
    pass


class IncorrectStateException(Exception):
    pass



def wrapped_post_future(future, state):

    async def web_token_response(request):
        future.set_result(await request.json())
        return web.json_response(data={"ok": True})

    return web_token_response

async def wait_for_post(starturl, previous={}, redirect_host="localhost", redirect_port=6767, redirect_path = "/", timeout=400, print_function= False, handle_signals=False):
    

    state = uuid.uuid4()
    #redirect_uri = quote(f"http://{redirect_host}:{redirect_port}{redirect_path}")

    params = {
        "state": state,
        "redirect_uri": f"http://{redirect_host}:{redirect_port}{redirect_path}",
        "scopes": previous.get("herre", {}).get("scopes", None),
        "name": previous.get("herre", {}).get("name", None),
        "client_id": previous.get("herre", {}).get("client_id", None),
    }

    print(params, previous)

    querystring = urlencode({key: value for key, value in params.items() if value != None})

    webbrowser.open_new(starturl + "?" + querystring)

    token_future = asyncio.get_event_loop().create_future()

    app = web.Application()
    cors = aiohttp_cors.setup(app, defaults={
    "*": aiohttp_cors.ResourceOptions(
            allow_headers="*",
        )
    })
    cors.add(app.router.add_post(redirect_path, wrapped_post_future(token_future, state)))


    webserver_task = asyncio.get_event_loop().create_task(web._run_app(app, host=redirect_host, port=redirect_port, print=print_function,handle_signals=handle_signals))
    done, pending = await asyncio.wait([token_future, webserver_task, asyncio.sleep(timeout)], return_when=asyncio.FIRST_COMPLETED)

    for tf in done:
        if tf == token_future:
            post_json = tf.result()
        else:
            post_json = None

    for task in pending:
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

    if not post_json: raise RetrieverException("No Post Data Received")
    return post_json


class FaktsRetriever:
    REDIRECT_HOST = "localhost"
    REDIRECT_PORT = 6767
    REDIRECT_PATH = "/"

    def __init__(self,*args, redirect_host= None, redirect_port = None, redirect_path = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.redirect_host = redirect_host or self.REDIRECT_HOST
        self.redirect_port = redirect_port or self.REDIRECT_PORT
        self.redirect_path = redirect_path or self.REDIRECT_PATH


    async def aretrieve(self, config: FaktsEndpoint, previous={}):
        post_data = await wait_for_post(config.url, previous = previous, redirect_host=self.redirect_host, redirect_port=self.redirect_port, redirect_path=self.redirect_path)
        return post_data

    def retrieve(self, config, as_task=True, **kwargs):
        return koil(self.aretrieve(config, **kwargs), as_task=as_task)






from PyQt5.QtGui import QTextLine
from fakts import Fakts
from qtpy import QtCore, QtWidgets
from koil.qt import FutureWrapper


class QtFakts(Fakts, QtWidgets.QWidget):
    loaded_signal = QtCore.Signal(bool)
    error_signal = QtCore.Signal(Exception)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,  **kwargs)
        self.layout = QtWidgets.QVBoxLayout()
        self.setWindowTitle("Retrieval Wizard")


        self.showf = FutureWrapper(koil = self.koil)
        self.showf.call.connect(self.handle_show)
        self.showfref = None

        self.hidef = FutureWrapper(koil = self.koil)
        self.hidef.call.connect(self.handle_hide)
        
        self.title = QtWidgets.QLabel("Konfig Wizard")
        self.title.setAlignment(QtCore.Qt.AlignCenter)
        self.layout.addWidget(self.title)
        self.subtitle = QtWidgets.QLabel("These are the grants we are able to use")
        self.layout.addWidget(self.subtitle)

        for grant in self.grants:
            grant_title = QtWidgets.QLabel(grant.__class__.__name__)
            self.layout.addWidget(grant_title)

        self.start_button = QtWidgets.QPushButton("Start")
        self.start_button.clicked.connect(self.on_start)
        self.layout.addWidget(self.start_button)


        self.setLayout(self.layout)


    def handle_show(self, ref, *args, **kwargs):
        self.show()
        self.showfref = ref

    def handle_hide(self, ref, *args, **kwargs):
        self.close()
        self.hidef.resolve.emit(ref, None)

    def on_start(self):
        self.showf.resolve.emit(self.showfref, None)


    async def aload(self):
        await self.showf.acall() # await user starts
        try:
            nana = await super().aload()
            self.loaded_signal.emit(True)
            await self.hidef.acall()
            return nana
        except Exception as e:
            self.error_signal.emit(e)
            await self.hidef.acall()
            raise e


    async def adelete(self):
        nana = await super().adelete()
        self.loaded_signal.emit(False)
        return nana


    



from typing import List
from pydantic import BaseModel
from socket import socket, AF_INET, SOCK_DGRAM, SOL_SOCKET, SO_BROADCAST
import asyncio
import json
from koil import koil
import logging


logger = logging.getLogger(__name__)


class FaktsEndpoint(BaseModel):
    url: str
    name: str



class BeaconProtocol(asyncio.DatagramProtocol):
    pass



class EndpointBeacon:
    BIND = ""
    BROADCAST_PORT = 45678
    BROADCAST_ADDRESS = "<broadcast>"
    MAGIC_PHRASE = "beacon-konfig"


    def __init__(self, broadcast_port=None, bind=None, magic_phrase = None, broadcast_adress = None,  advertised_endpoints: List[FaktsEndpoint] = [], interval=1) -> None:
        self.broadcast_port = broadcast_port or self.BROADCAST_PORT
        self.bind = bind or self.BIND
        self.magic_phrase = magic_phrase or self.MAGIC_PHRASE
        self.broadcast_adress = broadcast_adress or self.BROADCAST_ADDRESS

        self.advertised_endpoints = advertised_endpoints
        assert len(self.advertised_endpoints) > 0, "No config points provided"
        self.interval = interval


    def endpoint_to_message(self, config: FaktsEndpoint):
        return bytes(self.magic_phrase + json.dumps(config.dict()), "utf8")


    async def arun(self):

        s = socket(AF_INET, SOCK_DGRAM) #create UDP socket
        s.bind((self.bind, 0))
        s.setsockopt(SOL_SOCKET, SO_BROADCAST, 1) #this is a broadcast socket


        loop = asyncio.get_event_loop()
        transport, pr = await loop.create_datagram_endpoint(BeaconProtocol,sock=s)


        messages = [self.endpoint_to_message(config) for config in self.advertised_endpoints]       
        
        while True:
            for message in messages:
                transport.sendto(message, (self.broadcast_adress, self.broadcast_port))
                logger.info(f"Send Message {message}")

            await asyncio.sleep(self.interval)

    
    def run(self):
        return koil(self.arun())





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fakts',
 'fakts.beacon',
 'fakts.cli',
 'fakts.config',
 'fakts.grants',
 'fakts.grants.cli',
 'fakts.grants.meta',
 'fakts.grants.qt',
 'fakts.middleware',
 'fakts.middleware.environment']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'QtPy>=1.11.2,<2.0.0',
 'aiohttp_cors>=0.7.0,<0.8.0',
 'koil>=0.1.40,<0.2.0',
 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'fakts',
    'version': '0.1.16',
    'description': 'easy configuration discovery based on UDP broadcasts and konfig beacons (servers that redirect configs back to the client)',
    'long_description': None,
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
